#include <string.h>
#include <assert.h>
#include <ross.h>

#include "comb.h"
#include "mr-types.h"
#include "mr-extern.h"
#include "mr-config.h" 
#include "mr-event.h"

int 
datanodeGlobalToLocal( int gid )
{
  int lid, base, grp_id;
  // dragonfly specific
  base = num_tt_per_group + num_dn_per_group + num_tm_per_group + 1;
  grp_id = gid/base;
  // this one is tricky, because datanode is mapped
  // task_tracker, otherwise, you should add other lps
  lid = grp_id * num_dn_per_group + gid % base;
  return lid;
}

int 
datanodeLocalToGlobal( int lid )
{
  int gid, base, grp_id;
  // dragonfly specific
  base = num_tt_per_group + num_dn_per_group + num_tm_per_group + 1;
  grp_id = lid/num_dn_per_group;
  // this one is tricky, because datanode is mapped
  // task_tracker, otherwise, you should add other lps
  gid = grp_id * base + lid % num_dn_per_group + num_tt_per_group;
  return gid;
}

int
getLocalDatanode( int ttID )
{
  // based on this mapping 
  /* TaskTraker report, lp id is 0 */
  /*   TaskTraker report, lp id is 1 */
  /*   DataNode report, lp id is 2 */
  /*   DataNode report, lp id is 3 */
  /*   terminal 4  */
  /*   terminal 5  */
  /*   router 6  */

  return ttID + num_tt_per_group;

}

int
getLocalTaskTracker( int dnID )
{
  return dnID - num_dn_per_group;
}

// dragonfly specific
void 
init_block_map( tw_lpid ** block_map,
		int num_tt_per_group,
		int num_dn_per_group,
		int num_tm_per_group,
		int num_routers )
{
  int i, j, base;
  // place all copies, round robin
  base = num_tt_per_group + num_dn_per_group + num_tm_per_group + 1;
  for (i=0;i<N_BLOCK;i++)
    for (j=0;j<N_REPLICA;j++)
      {
	block_map[i][j] = (i*base + j*base + num_tt_per_group ) % (num_routers * base) ;
	// second copy modification, same rack (group)
	if (j==1)
	  block_map[i][j] = block_map[i][0]+1;
      }

  /* for (i=0;i<N_BLOCK;i++) */
  /*   { */
  /*     for (j=0;j<N_REPLICA;j++) */
  /* 	printf( " %d", block_map[i][j] ); */
  /*     printf("\n"); */
  /*   } */
}

void 
datanode_init( datanode_state* ns,  
	       tw_lp* lp )
{

    printf("DataNode report, lp id is %d\n",lp->gid);
    
    int i;
    if (lp->gid == NAMENODE_POS) // namenode
      {
	ns->nn_state = (namenode_state *)malloc(sizeof(namenode_state));
	(ns->nn_state)->block_map = (tw_lpid **)malloc(sizeof(tw_lpid *) * N_BLOCK);
	for (i=0;i<N_BLOCK;i++)
	  (ns->nn_state)->block_map[i] = (tw_lpid *)malloc(sizeof(tw_lpid) * N_REPLICA);
	// initialize block_map, it should be based on config file in future
	// it is purely computed by now, and evenly distributed
	init_block_map( (ns->nn_state)->block_map,
			num_tt_per_group,
			num_dn_per_group,
			num_tm_per_group,
			num_routers);
      }
    else // datanode
      {
	ns->nn_state = NULL;
      }

    ns->read_channel_next_available_time = 0;
    ns->local_IO_channel_next_available_time = 0;
    ns->query_msg_count = 0;

    return;
}

void datanode_eventHandler(datanode_state* ns, tw_bf* b, mr_msg* m, tw_lp* lp)
{
   switch (m->mr_event_type)
    {
    case MR_GET_INTER_DATA:
      handle_get_inter_data_event(ns, b, m, lp);
      break;      
    case MR_PROCESS_GET_DATA_BLOCK:
      handle_process_get_data_block_event(ns, b, m, lp);
      break;      
    case MR_RECV_BLOCK_POS_QUERY:
      handle_recv_block_pos_query_event(ns, b, m, lp);
      break;      
    case MR_REQ:
      handle_req_event(ns, b, m, lp);
      break;
    case MR_ACK:
      handle_ack_event(ns, b, m, lp);
      break;
    case MR_KICKOFF:
      handle_kickoff_event(ns, b, m, lp);
      break;
    case MR_LOCAL:
      handle_local_event(ns, b, m, lp);
      break;
    default:
      printf("\n Invalid message type %d ", m->mr_event_type);
      assert(0);
      break;
    }
}

void datanode_eventHandler_rc(datanode_state* ns, tw_bf* b, mr_msg* m, tw_lp* lp)
{
    switch (m->mr_event_type)
    {
    case MR_GET_INTER_DATA:
      handle_get_inter_data_event_rc(ns, b, m, lp);
      break;      
    case MR_PROCESS_GET_DATA_BLOCK:
      handle_process_get_data_block_event_rc(ns, b, m, lp);
      break;      
    case MR_RECV_BLOCK_POS_QUERY:
      handle_recv_block_pos_query_event_rc(ns, b, m, lp);
      break;            
    /* case MR_REQ: */
    /*   handle_req_rev_event(ns, b, m, lp); */
    /*   break; */
    /* case MR_ACK: */
    /*   handle_ack_rev_event(ns, b, m, lp); */
    /*   break; */
    /* case MR_KICKOFF: */
    /*   handle_kickoff_rev_event(ns, b, m, lp); */
    /*   break; */
    /* case MR_LOCAL: */
    /*   handle_local_rev_event(ns, b, m, lp);     */
    /*   break; */
    default:
      assert(0);
      break;
    }
    
    return;
}

/* convert ns to seconds */
static tw_stime ns_to_s(tw_stime ns)
{
    return(ns / (1000.0 * 1000.0 * 1000.0));
}

/* convert seconds to ns */
static tw_stime s_to_ns(tw_stime ns)
{
    return(ns * (1000.0 * 1000.0 * 1000.0));
}

void datanode_finish(datanode_state* ns, tw_lp* lp)
{
  /* int i; */
  /* if (lp->gid == NAMENODE_POS) // namenode */
  /*   { */
  /*     for (i=0;i<N_BLOCK;i++) */
  /* 	free ((ns->nn_state)->block_map[i]); */
      
  /*     free ((ns->nn_state)->block_map); */
  /*     free(ns->nn_state); */
  /*   } */

  //printf("Data Node %d finished\n",lp->gid);
    /* printf("server %llu recvd %d bytes in %f seconds, %f MiB/s sent_count %d recvd_count %d local_count %d \n", (unsigned long long)lp->gid, PAYLOAD_SZ*ns->msg_recvd_count, ns_to_s((tw_now(lp)-ns->start_ts)),  */
    /*     ((double)(PAYLOAD_SZ*NUM_REQS)/(double)(1024*1024)/ns_to_s(tw_now(lp)-ns->start_ts)), ns->msg_sent_count, ns->msg_recvd_count, ns->local_recvd_count); */
    return;
}
