#ifndef MR_TYPES_H
#define MR_TYPES_H

#include "mr-config.h" 
#include "mr-event.h"

/*
 * define helper struct
 */
typedef struct 
{
  unsigned long long range_start;
  unsigned long long range_end;
  int num_splits;
}mrJob;

typedef enum
  {
    MAP,
    REDUCE
  }mr_task_T;

typedef struct
{
  mr_task_T taskType; 
  int neededBlock[MAX_BLOCK_PER_TASK];
}mrTask;

typedef struct
{
  tw_stime beatTime;
  int numFinishedTask;
}heartBeatMsgT;

typedef struct
{
  int blockSize;
  int blockID;
}dataMsgT;

typedef struct
{
  int jobSize;
  
}metaMsgT;

typedef union 
{
  dataMsgT dataMsg;
  metaMsgT metaMsg;
  heartBeatMsgT heartMsg; 
}coreMsgT;

typedef struct
{
  int dnPos[N_REPLICA];
}mrBlock;


/*
 * unify all messages in MapReduce systems
 */
typedef struct
{
  enum mr_event mr_event_type;
  heartBeatMsgT heart_beat_msg;
  mrTask mr_task;
  mrBlock mr_block;
  tw_lpid src;
  int incremented_flag;

  coreMsgT coreMsg;
}mr_msg;

/*
 * define states for all lp types
 *  job_tracker, task_tracker,
 * namenode, datanode, and more to come
 */
typedef struct
{
  int * slotMap; 
  mrJob mr_job[MAX_NUM_JOB];
  tw_lpid ** job_schedule;
  int map_task_finish_count;

  int msg_sent_count;       /* requests sent */
  int msg_recvd_count;      /* requests recvd */
  int local_recvd_count;    /* number of local messages received */
  tw_stime start_ts;        /* time that we started sending requests */
}job_tracker_state;

typedef struct
{
  job_tracker_state * jt_state; // this is for job tracker
  tw_lpid this_terminal_lpid;   // used for tracking msg in this node
  int reduce_task_finish_count;
  int counter_1;

  int msg_sent_count;       /* requests sent */
  int msg_recvd_count;      /* requests recvd */
  int local_recvd_count;    /* number of local messages received */
  tw_stime start_ts;        /* time that we started sending requests */
}task_tracker_state;

typedef struct
{
  tw_lpid ** block_map;

  int msg_sent_count;       /* requests sent */
  int msg_recvd_count;      /* requests recvd */
  int local_recvd_count;    /* number of local messages received */
  tw_stime start_ts;        /* time that we started sending requests */
}namenode_state;

typedef struct
{
  namenode_state * nn_state; // temporary 
  tw_stime read_channel_next_available_time;
  tw_stime local_IO_channel_next_available_time;

  int query_msg_count;

  int msg_sent_count;       /* requests sent */
  int msg_recvd_count;      /* requests recvd */
  int local_recvd_count;    /* number of local messages received */
  tw_stime start_ts;        /* time that we started sending requests */
}datanode_state;

#endif
