#include <string.h>
#include <assert.h>
#include <ross.h>

#include "comb.h"
#include "mr-types.h"
#include "mr-extern.h"
#include "mr-config.h" 
#include "mr-event.h"

int net_id = 0;
int num_routers = 0;
int num_task_tracker = 0;
int num_datanode = 0;
int offset = 2;
int num_tt_per_group = 0;
int num_dn_per_group = 0;
int num_tm_per_group = 0;
int NAMENODE_POS = 3;

tw_lptype mr_lps[] =
  {
    {
      (init_f) task_tracker_init,
      (event_f) task_tracker_eventHandler,
      (revent_f) task_tracker_eventHandler_rc,
      (final_f) task_tracker_finish,
      (map_f) codes_mapping,
      sizeof( task_tracker_state )
    },
    {
      (init_f) datanode_init,
      (event_f) datanode_eventHandler,
      (revent_f) datanode_eventHandler_rc,
      (final_f) datanode_finish,
      (map_f) codes_mapping,
      sizeof( datanode_state )
    },
    {0},
  };

const tw_optdef app_opt [] =
{
	TWOPT_GROUP("MapReduce Model" ),
	TWOPT_END()
};

/*
 * This function is based the premises
 * that number of task trakcer deamon 
 * and datanode deamon are equal in each node,
 * and they are roundRobinly mapped
 * also num_tt_per_group and
 * num_dn_per_group are computed 
 */
int isInSameNode(int a, int b)
{
  // if not in same group then definitely
  // not same node
  int base = num_tt_per_group + num_dn_per_group + num_tm_per_group + 1;
  if (a/base != b/base)
    {
      //printf(" %d and %d returned in first if\n", a, b);
      return 0;
    }
  if (( (a%base) % num_tm_per_group) != ( (b%base) % num_tm_per_group))
    {
      //printf(" a %d and b %d here\n",  (a%base) % num_tm_per_group , (b%base) % num_tm_per_group );
      return 0;
    }
  else
    return 1;
}

int main( int argc, char **argv )
{
    int nprocs;
    int rank;

    tw_opt_add(app_opt);
    tw_init(&argc, &argv);

    if(argc < 2)
    {
	printf("\n Usage: mpirun <args> --sync=2/3 mapping_file_name.conf (optional --nkp) ");
	MPI_Finalize();
	return 0;
    }

    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &nprocs);
  
    configuration_load(argv[2], MPI_COMM_WORLD, &config);
    net_id=model_net_set_params();

    lp_type_register("task_tracker", &mr_lps[0]);
    lp_type_register("datanode", &mr_lps[1]);

    codes_mapping_setup();
    
    num_task_tracker = codes_mapping_get_group_reps("MODELNET_GRP") * codes_mapping_get_lp_count("MODELNET_GRP", "task_tracker");
    num_datanode = codes_mapping_get_group_reps("MODELNET_GRP") * codes_mapping_get_lp_count("MODELNET_GRP", "datanode");

    num_tt_per_group = codes_mapping_get_lp_count("MODELNET_GRP", "task_tracker");
    num_dn_per_group = codes_mapping_get_lp_count("MODELNET_GRP", "datanode");
    num_tm_per_group = codes_mapping_get_lp_count("MODELNET_GRP", "modelnet_dragonfly");

    printf("\nNumber of Ttracker is %d and Datanode is %d\n", num_task_tracker, num_datanode);

    if(net_id == DRAGONFLY)
    {
	num_routers = codes_mapping_get_group_reps("MODELNET_GRP") * codes_mapping_get_lp_count("MODELNET_GRP", "dragonfly_router"); 
    }

    NAMENODE_POS = num_tm_per_group;
    g_tw_events_per_pe = 100000000;
    tw_run();

    tw_end();
    return 0;
}
