#include <string.h>
#include <assert.h>
#include <ross.h>

#include "comb.h"
#include "mr-types.h"
#include "mr-extern.h"
#include "mr-config.h" 
#include "mr-event.h"

void handle_heartbeat_event(
    task_tracker_state * ns,
    tw_bf * b,
    mr_msg * m,
    tw_lp * lp)
{
  tw_stime interval = HEART_BEAT_INTERVAL;
  tw_event *e;
  mr_msg *msg;

  int dest_id = JOB_TRACKER_POS;
  mr_msg * m_local = malloc(sizeof(mr_msg));
  mr_msg * m_remote = malloc(sizeof(mr_msg));

  m_local->mr_event_type = MR_LOCAL;
  m_local->src = lp->gid;

  // testing for now, it should check the real finished tasks
  // and then send
  (m_local->heart_beat_msg).numFinishedTask = 1;

  memcpy(m_remote, m_local, sizeof(mr_msg));
  m_remote->mr_event_type = MR_RECV_HEART_BEAT;

  printf("time now is %lf\n",tw_now(lp));

  // recursively schedule self 
  if ( tw_now(lp) + interval < g_tw_ts_end)
    {
      e = codes_event_new(lp->gid, interval, lp);
      msg = tw_event_data(e);
      msg->mr_event_type = MR_HEART_BEAT;
      tw_event_send(e);
    }

  // and send beat msg to job tracker
#if DEBUG_HELPER
  printf("lp %d start heart beat and is same group with %d, answer %d\n",
	 lp->gid,JOB_TRACKER_POS,isInSameNode(lp->gid,JOB_TRACKER_POS));
#endif

  // if local send directly
  if (isInSameNode(lp->gid,JOB_TRACKER_POS))
    {
      e = codes_event_new(dest_id, LOCAL_MSG_TRANS_INTERVAL, lp);
      msg = tw_event_data(e);
      msg->mr_event_type = MR_RECV_HEART_BEAT;
      tw_event_send(e);
    }
  else// otherwise send through network
    {
      model_net_event( net_id, 
		       "test", 
		       dest_id, 
		       HEART_BEAT_MSG_SIZE, 
		       sizeof(mr_msg), 
		       (const void*)m_remote, 
		       sizeof(mr_msg), 
		       (const void*)m_local, 
		       lp );
    }

}

void handle_heartbeat_recv_event(
    task_tracker_state * ns,
    tw_bf * b,
    mr_msg * m,
    tw_lp * lp)
{
  // only job tracker deal with this event
  printf("lp %d received heart beat\n",lp->gid);
  int lid;
  lid = taskTrackerGlobalToLocal(m->src);
  printf("gid is %d and lid is %d\n",m->src,lid);
  // 
  if ( ((ns->jt_state)->slotMap[lid]==0) && ((m->heart_beat_msg).numFinishedTask) )
    printf("Error, trying to release non-existing slot in node %d\n",m->src);
  else
    (ns->jt_state)->slotMap[lid] -= (m->heart_beat_msg).numFinishedTask;

  //outputSlotMap((ns->jt_state)->slotMap);
  
}

void 
handle_get_data_block_event( task_tracker_state * ns,
			     tw_bf * b,
			     mr_msg * m,
			     tw_lp * lp)
{
  //printf("Get data block event %d\n",lp->gid);
  // after a long pause, computation
  // start reduce task

  tw_event *e;
  mr_msg *msg;
  tw_stime interval;

  // currently assuming no local read contention
  interval = LOCAL_MSG_TRANS_INTERVAL;
  //printf("interval is %lf",interval);

  // request should go to local datanode daemon
  int dest_id = getLocalDatanode(lp->gid);

  e = codes_event_new(dest_id, interval, lp);
  msg = tw_event_data(e);
  msg->src = lp->gid;
  msg->mr_event_type = MR_PROCESS_GET_DATA_BLOCK;
  tw_event_send(e);

}


void 
handle_get_data_block_event_rc( task_tracker_state * ns,
				tw_bf * b,
				mr_msg * m,
				tw_lp * lp)
{
}

void 
handle_get_data_block_back_event( task_tracker_state * ns,
				  tw_bf * b,
				  mr_msg * m,
				  tw_lp * lp)
{
  ns->counter_1++;
  printf("Now ACK pos query count is %d\n",ns->counter_1);


  printf(" \n*********** \nThis is %d and time now is %lf, get data block back\n", lp->gid, tw_now(lp));
  printf("Get data block event %d\n",lp->gid);
  // after a long pause, computation
  // start reduce task

  tw_event *e;
  mr_msg *msg;
  tw_stime interval;

  // currently assuming no local read contention
  interval = LOCAL_MSG_TRANS_INTERVAL;
  //printf("interval is %lf",interval);

  // request should go to local datanode daemon

  e = codes_event_new(lp->gid, interval, lp);
  msg = tw_event_data(e);
  msg->src = lp->gid;
  msg->mr_event_type = MR_START_MAP_TASK;
  tw_event_send(e);  
  
}

void 
handle_start_map_task_event( task_tracker_state * ns,
			     tw_bf * b,
			     mr_msg * m,
			     tw_lp * lp)
{
  tw_event *e;
  mr_msg *msg;
  tw_stime interval;

  interval = MAP_COMPUTE_TIME;
  //printf("interval is %lf",interval);

  // after computation, each map task report to job tracker
  // will add inbalanced computation in future
  e = codes_event_new(JOB_TRACKER_POS, interval, lp);
  msg = tw_event_data(e);
  msg->src = lp->gid;
  msg->mr_event_type = MR_FINISH_MAP_TASK;
  tw_event_send(e);  
  
  // start reduce task in the middle, with a latency ration
  // defined in config.h

  interval = MAP_COMPUTE_TIME * REDUCE_TASK_LATENCY_RATIO;
  //printf("interval is %lf",interval);

  e = codes_event_new(lp->gid, interval, lp);
  msg = tw_event_data(e);
  msg->src = lp->gid;
  msg->mr_event_type = MR_START_REDUCE_TASK;
  tw_event_send(e);  
  
}

void 
handle_start_map_task_event_rc( task_tracker_state * ns,
				tw_bf * b,
				mr_msg * m,
				tw_lp * lp)
{
}

void 
handle_start_reduce_task_event( task_tracker_state * ns,
				tw_bf * b,
				mr_msg * m,
				tw_lp * lp)
{
  printf("\t \n *** LP %d start start reduce task\n",lp->gid);
  tw_event *e;
  mr_msg *msg;
  tw_stime interval;
  tw_lpid dest_id;

  mr_msg * m_local = malloc(sizeof(mr_msg));
  mr_msg * m_remote = malloc(sizeof(mr_msg));

  m_local->mr_event_type = MR_LOCAL;
  m_local->src = lp->gid;

  memcpy(m_remote, m_local, sizeof(mr_msg));
  m_remote->mr_event_type = MR_GET_INTER_DATA;

  // need to first retrieve the intermediate results
  // from mappers

  /* tw_lpid * targetNode; */
  /* int * numNode; */

  int jobSize = NUM_TASK_PER_JOB;

  /* numNode=(int *)malloc(sizeof(int)); */
  /* targetNode = (tw_lpid *)malloc(sizeof(tw_lpid)*jobSize); */
  int numNode = -1; 
  tw_lpid targetNode[NUM_TASK_PER_JOB];
  
  int i, j;
  
  // scheduling module
  scheduleReduceTask( &targetNode,
		      &numNode );
  // schedule end, targeted nodes is in targetNode
  
#ifdef DEBUG_SCHEDULER
  printf("Number of nodes is %d\n",*numNode);
#endif
  
  for (i=0;i<jobSize;i++)
    {
      
#ifdef DEBUG_SCHEDULER
      printf("Scheduled node is %d\n",targetNode[i]);
#endif
      // put scheduled node id to the
      // job schedule table, (in jt state)
      // first change to global id
      // put -1 if out of range

      // send request to datanode
      // if local send local
      // if remote send through network
      dest_id = datanodeLocalToGlobal(targetNode[i]);
      //printf("\t Round %d Dest ID is %d\n",i,dest_id);

      // if local send directly
      if (isInSameNode(lp->gid,dest_id))
	{
	  e = codes_event_new(dest_id, LOCAL_MSG_TRANS_INTERVAL, lp);
	  msg = tw_event_data(e);
	  msg->src = lp->gid;
	  msg->mr_event_type =  MR_GET_INTER_DATA;
	  tw_event_send(e);
	}
      else// otherwise send through network
	{
	  model_net_event( net_id,
			   "test",
			   dest_id,
			   SCHEDULE_META_MSG_SIZE,
			   sizeof(mr_msg),
			   (const void*)m_remote,
			   sizeof(mr_msg),
			   (const void*)m_local,
			   lp );
	}      
    }
  
  /* free(numNode); */
  /* free(targetNode); */
}

void 
handle_start_reduce_task_event_rc( task_tracker_state * ns,
				   tw_bf * b,
				   mr_msg * m,
				   tw_lp * lp)
{
}


void 
handle_finish_reduce_task_event( task_tracker_state * ns,
				 tw_bf * b,
				 mr_msg * m,
				 tw_lp * lp)
{
  printf("\t\n This is lp %d and finished reduce task at %lf\n",lp->gid,tw_now(lp));
  
}

void 
handle_finish_reduce_task_event_rc( task_tracker_state * ns,
				    tw_bf * b,
				    mr_msg * m,
				    tw_lp * lp)
{
  printf("\t\n This is lp %d and finished reduce task at %lf\n",lp->gid,tw_now(lp));
  
}


void 
handle_finish_map_task_event_rc( task_tracker_state * ns,
				 tw_bf * b,
				 mr_msg * m,
				 tw_lp * lp)
{
  
  //printf("\t\n This is lp %d and finished map task at %lf\n",lp->gid,tw_now(lp));
  
}

void 
handle_finish_map_task_event( task_tracker_state * ns,
			      tw_bf * b,
			      mr_msg * m,
			      tw_lp * lp)
{
  (ns->jt_state)->map_task_finish_count++;
  printf(" NOW finished map tasks is %d\n",(ns->jt_state)->map_task_finish_count);

  if ( (ns->jt_state)->map_task_finish_count == MAX_TASK_PER_JOB )
    {
      printf("\n\n\t ********************************* \n");
      printf("\t\n This is lp %d and finished map task at %lf\n\n",lp->gid,tw_now(lp));
      printf("\n\t ********************************* \n");
      (ns->jt_state)->map_task_finish_count = 0;
    }
  
}


void 
handle_get_data_block_back_event_rc( task_tracker_state * ns,
				     tw_bf * b,
				     mr_msg * m,
				     tw_lp * lp)
{
}

void handle_assign_map_task_event_rc(
    task_tracker_state * ns,
    tw_bf * b,
    mr_msg * m,
    tw_lp * lp)
{
}

void handle_heartbeat_event_rc(
    task_tracker_state * ns,
    tw_bf * b,
    mr_msg * m,
    tw_lp * lp)
{
}

void 
handle_recv_map_task_event_rc( task_tracker_state * ns,
			       tw_bf * b,
			       mr_msg * m,
			       tw_lp * lp)
{
}

void
handle_get_inter_data_back_event_rc( task_tracker_state * ns,
				     tw_bf * b,
				     mr_msg * m,
				     tw_lp * lp)
{
}

void
handle_get_inter_data_back_event( task_tracker_state * ns,
				  tw_bf * b,
				  mr_msg * m,
				  tw_lp * lp)
{
  printf("\t\n This is task tracker %d and get *** %%% $$$ DATA BACK from %d time now is %lf\n",lp->gid,m->src,tw_now(lp));
  // this is the end of shuffle phase,
  // if this works, we can schedule a simple reduce task and call it done
  // note that, this process should collect all intermediate data and
  // then start to schedule reduce task

  tw_event *e;
  mr_msg *msg;
  tw_stime interval = REDUCE_COMPUTE_TIME;
  int i;

  ns->reduce_task_finish_count++;

  if (ns->reduce_task_finish_count==NUM_TASK_PER_JOB)
    {
      e = codes_event_new(lp->gid, interval, lp);
      msg = tw_event_data(e);
      msg->src = lp->gid;
      msg->mr_event_type = MR_FINISH_REDUCE_TASK;
      tw_event_send(e);
      ns->reduce_task_finish_count = 0;
    }

}  

void 
handle_schedule_event_rc( task_tracker_state * ns,
			  tw_bf * b,
			  mr_msg * m,
			  tw_lp * lp)
{
}

void handle_heartbeat_recv_event_rc(
    task_tracker_state * ns,
    tw_bf * b,
    mr_msg * m,
    tw_lp * lp)
{
}

void 
handle_kickoff_event( task_tracker_state * ns,
		      tw_bf * b,
		      mr_msg * m,
		      tw_lp * lp)
{
  tw_event *e;
  mr_msg *msg;

  tw_stime interval = LOCAL_MSG_TRANS_INTERVAL;
  
  // local event
  if (lp->gid == JOB_TRACKER_POS)
    {
      interval = LOCAL_MSG_TRANS_INTERVAL;
      e = codes_event_new(lp->gid, interval, lp);
      msg = tw_event_data(e);

      msg->mr_event_type = MR_SCHEDULE;
      tw_event_send(e);
    }  
}

void 
handle_schedule_event( task_tracker_state * ns,
		       tw_bf * b,
		       mr_msg * m,
		       tw_lp * lp)
{
  // only job tracker deal with this event
  // schedule decision is based on slotMap and the job
  if (lp->gid==JOB_TRACKER_POS)
    {
      tw_event *e;
      mr_msg *msg;
      
      int i, j;

      for (j=0;j<MAX_NUM_JOB;j++)
	{
	  // scheduling module
	  scheduleMapTask( (ns->jt_state)->slotMap,
	  		   (ns->jt_state)->mr_job[j],
	  		   (ns->jt_state)->job_schedule,
	  		   j );
	  // schedule end, targeted nodes is in targetNode
	}

      //outputScheduleMap((ns->jt_state)->job_schedule);

      e = codes_event_new(lp->gid, LOCAL_MSG_TRANS_INTERVAL, lp);
      msg = tw_event_data(e);
      msg->mr_event_type = MR_ASSIGN_MAP_TASK;
      tw_event_send(e);

    }
  else
    printf("\t \nSchedule event happend at task tracker, please check receiver IP.\n ");
}

void 
handle_assign_map_task_event( task_tracker_state * ns,
			      tw_bf * b,
			      mr_msg * m,
			      tw_lp * lp)
{
  tw_event *e;
  mr_msg *msg;
  tw_stime interval;
  int dest_id;

  mr_msg * m_local = malloc(sizeof(mr_msg));
  mr_msg * m_remote = malloc(sizeof(mr_msg));

  m_local->mr_event_type = MR_LOCAL;
  m_local->src = lp->gid;

  // temp assign one associated block 
  // to each task
  m_local->mr_task.neededBlock[0] = 1;

  memcpy(m_remote, m_local, sizeof(mr_msg));
  m_remote->mr_event_type = MR_RECV_MAP_TASK;

  int i, j;
  
#ifdef DEBUG_SCHEDULER
  printf("Number of nodes is %d\n",*numNode);
#endif
  
  //outputScheduleMap((ns->jt_state)->job_schedule);

  for (j=0;j<ACTIVE_JOB;j++)
    for (i=0;i<MAX_TASK_PER_JOB;i++)
      {
	dest_id = (ns->jt_state)->job_schedule[j][i];
	//printf("Dest id is %d\n",dest_id);
	//dest_id = 1;
	// check valid assignment
	if (dest_id>=0)
	  {
	    /* ns->counter_1++; */
	    /* printf("Now send out block pos query count is %d\n",ns->counter_1); */

	    // if local send directly
	    if (isInSameNode(lp->gid,dest_id))
	      {
		//printf("number 1 is %d\n",i);
		e = codes_event_new(dest_id, LOCAL_MSG_TRANS_INTERVAL, lp);
		msg = tw_event_data(e);
		msg->mr_event_type = MR_RECV_MAP_TASK;
		// round robin assign one associated block 
		// to each task
		msg->mr_task.neededBlock[0] = dest_id % N_BLOCK;
		
		/* // dont send to job tracker */
		/* if (dest_id != JOB_TRACKER_POS) */
		  tw_event_send(e);
	
	      }
	    else// otherwise send through network
	      {
		//printf("number 2 is %d\n",i);
		// round robin assign one associated block 
		// to each task
		m_remote->mr_task.neededBlock[0] = dest_id % N_BLOCK;

		// shouldn't all flush to net at once
		

		/* // dont send to job tracker */
		/* if (dest_id != JOB_TRACKER_POS) */
		  model_net_event( net_id, 
				   "test", 
				   dest_id, 
				   SCHEDULE_META_MSG_SIZE, 
				   sizeof(mr_msg), 
				   (const void*)m_remote, 
				   sizeof(mr_msg), 
				   (const void*)m_local, 
				   lp );
	      }
	  }
      }
}

void 
handle_recv_map_task_event( task_tracker_state * ns,
			    tw_bf * b,
			    mr_msg * m,
			    tw_lp * lp)
{
  /* ns->counter_1++; */
  /* printf("Now send out block pos query count is %d\n",ns->counter_1); */

  /* printf("In recv map task event ************************ %d \n",lp->gid); */
  /* printf("lp %d block %d\n",lp->id,m->mr_task.neededBlock[0]); */

  tw_event *e;
  mr_msg *msg;
  tw_stime interval;

  e = codes_event_new(lp->gid, LOCAL_MSG_TRANS_INTERVAL, lp);
  //e = codes_event_new(1, LOCAL_MSG_TRANS_INTERVAL, lp);
  msg = tw_event_data(e);
  msg->src = lp->gid;
  memcpy( msg->mr_task.neededBlock,
  	  m->mr_task.neededBlock,
  	  sizeof(int)*MAX_BLOCK_PER_TASK );
  //printf("needed block is %d\n",msg->mr_task.neededBlock[0]);
  msg->mr_event_type = MR_BLOCK_POS_QUERY;
  tw_event_send(e);

}

void 
handle_block_pos_query_event( task_tracker_state * ns,
			      tw_bf * b,
			      mr_msg * m,
			      tw_lp * lp)
{
  // first ask namenode for block position
  // now task tracker get an task
  // ask namenode for block position

  /* ns->counter_1++; */
  /* printf("Now send out block pos query count is %d my id is %d\n",ns->counter_1,m->src); */

  //printf("In pos query and received needed block is %d\n",m->mr_task.neededBlock[0]);

  tw_event *e;
  mr_msg *msg;
  tw_stime interval;
  // query send to namenode
  int dest_id = NAMENODE_POS;

  mr_msg * m_local = malloc(sizeof(mr_msg));
  mr_msg * m_remote = malloc(sizeof(mr_msg));

  m_local->mr_event_type = MR_LOCAL;
  m_local->src = lp->gid;

  memcpy(m_remote, m_local, sizeof(mr_msg));
  m_remote->mr_event_type = MR_RECV_BLOCK_POS_QUERY;

  // if local send directly
  if (isInSameNode(lp->gid,dest_id))
    {
      e = codes_event_new(dest_id, LOCAL_MSG_TRANS_INTERVAL, lp);
      msg = tw_event_data(e);
      msg->src = lp->gid;
      memcpy( &(msg->mr_task),
  	      &(m->mr_task),
  	      sizeof(mrTask) );

      msg->mr_event_type = MR_RECV_BLOCK_POS_QUERY;
      tw_event_send(e);
    }
  else// otherwise send through network
    {
      memcpy( &(m_remote->mr_task),
      	      &(m->mr_task),
      	      sizeof(mrTask) );

      model_net_event( net_id,
      		       "test",
      		       dest_id,
      		       QUERY_MSG_SIZE,
      		       sizeof(mr_msg),
      		       (const void*)m_remote,
      		       sizeof(mr_msg),
      		       (const void*)m_local,
      		       lp );
    }
}

void 
handle_ack_block_pos_query_event( task_tracker_state * ns,
				  tw_bf * b,
				  mr_msg * m,
				  tw_lp * lp)
{
  /* ns->counter_1++; */
  /* printf("Now ACK pos query count is %d\n",ns->counter_1); */

  printf("In ack block pos query and this is %d, \n",
	 lp->gid);
  int i;
  for (i=0;i<N_REPLICA;i++)
    printf("block in datanode %d\n",
	   m->mr_block.dnPos[i]);

  // the available block location is provided in 
  // dnPos, and here, we only read from local
  // assuming that block is local, and this will
  // change in future

  tw_event *e;
  mr_msg *msg;
  tw_stime interval;

  e = codes_event_new(lp->gid, LOCAL_MSG_TRANS_INTERVAL, lp);
  msg = tw_event_data(e);
  msg->src = lp->gid;
  msg->mr_event_type = MR_GET_DATA_BLOCK;
  tw_event_send(e);

}


void handle_block_pos_query_event_rc(
		task_tracker_state * ns,
		tw_bf * b,
		mr_msg * m,
		tw_lp * lp)
{
}

void handle_ack_block_pos_query_event_rc(
		task_tracker_state * ns,
		tw_bf * b,
		mr_msg * m,
		tw_lp * lp)
{
}


void handle_local_event(
		task_tracker_state * ns,
		tw_bf * b,
		mr_msg * m,
		tw_lp * lp)
{
  ns->local_recvd_count++;
}

void handle_local_rev_event(
	       task_tracker_state * ns,
	       tw_bf * b,
	       mr_msg * m,
	       tw_lp * lp)
{
    ns->local_recvd_count--;
}
/* reverse handler for req event */
void handle_req_rev_event(
    task_tracker_state * ns,
    tw_bf * b,
    mr_msg * m,
    tw_lp * lp)
{
    ns->msg_recvd_count--;
    model_net_event_rc(net_id, lp, PAYLOAD_SZ);

    return;
}

/* reverse handler for kickoff */
void handle_kickoff_rev_event(
    task_tracker_state * ns,
    tw_bf * b,
    mr_msg * m,
    tw_lp * lp)
{
    ns->msg_sent_count--;
    model_net_event_rc(net_id, lp, PAYLOAD_SZ);

    return;
}

/* reverse handler for ack*/
void handle_ack_rev_event(
    task_tracker_state * ns,
    tw_bf * b,
    mr_msg * m,
    tw_lp * lp)
{
    if(m->incremented_flag)
    {
        model_net_event_rc(net_id, lp, PAYLOAD_SZ);
        ns->msg_sent_count--;
    }
    return;
}

/* handle recving ack */
void handle_ack_event(
    task_tracker_state * ns,
    tw_bf * b,
    mr_msg * m,
    tw_lp * lp)
{
    mr_msg * m_local = malloc(sizeof(mr_msg));
    mr_msg * m_remote = malloc(sizeof(mr_msg));

    m_local->mr_event_type = MR_LOCAL;
    m_local->src = lp->gid;

    memcpy(m_remote, m_local, sizeof(mr_msg));
    m_remote->mr_event_type = MR_REQ;

//    printf("handle_ack_event(), lp %llu.\n", (unsigned long long)lp->gid);

    /* safety check that this request got to the right server */
//    printf("\n m->src %d lp->gid %d ", m->src, lp->gid);
    int opt_offset = 0;
    if(net_id == DRAGONFLY && lp->gid % 5)
	 opt_offset = 3;

    assert(m->src == (lp->gid + offset + opt_offset)%(num_task_tracker*2 + num_routers));

    if(ns->msg_sent_count < NUM_REQS)
    {
        /* send another request */
	model_net_event(net_id, "test", m->src, PAYLOAD_SZ, sizeof(mr_msg), (const void*)m_remote, sizeof(mr_msg), (const void*)m_local, lp);
        ns->msg_sent_count++;
        m->incremented_flag = 1;
    }
    else
    {
        m->incremented_flag = 0;
    }

    return;
}

/* handle receiving request */
void handle_req_event(
    task_tracker_state * ns,
    tw_bf * b,
    mr_msg * m,
    tw_lp * lp)
{
    mr_msg * m_local = malloc(sizeof(mr_msg));
    mr_msg * m_remote = malloc(sizeof(mr_msg));

    m_local->mr_event_type = MR_LOCAL;
    m_local->src = lp->gid;

    memcpy(m_remote, m_local, sizeof(mr_msg));
    m_remote->mr_event_type = MR_ACK;
    //printf("handle_req_event(), lp %llu src %llu .\n", (unsigned long long)lp->gid, (unsigned long long) m->src);

    /* safety check that this request got to the right server */
    printf("\n m->src %d lp->gid %d ", m->src, lp->gid);
    int opt_offset = 0;
    if(net_id == DRAGONFLY && (m->src % 5))
	  opt_offset = 3; /* optional offset due to dragonfly mapping */

    printf("In act event, offset is %d\n",offset);
    
    //assert(lp->gid == (m->src + offset + opt_offset)%(num_task_tracker*2 + num_routers));
    ns->msg_recvd_count++;

    /* send ack back */
    /* simulated payload of 1 MiB */
    /* also trigger a local event for completion of payload msg */
    /* remote host will get an ack event */
   
   // mm Q: What should be the size of an ack message? may be a few bytes? or larger..? 
    //model_net_event(net_id, "test", m->src, PAYLOAD_SZ, sizeof(mr_msg), (const void*)m_remote, sizeof(mr_msg), (const void*)m_local, lp);

//    printf("\n Sending ack to LP %d %d ", m->src, m_remote->src);
    return;
}
