#include <string.h>
#include <assert.h>
#include <ross.h>

#include "comb.h"
#include "mr-types.h"
#include "mr-extern.h"
#include "mr-config.h" 
#include "mr-event.h"

void scheduleMapTask( int * slotMap,
		      mrJob mr_job,
		      tw_lpid ** targetNode,
		      int jobPos )
{
  // use fair schedule for now
  int i;
  
  for (i=0;i<MAX_TASK_PER_JOB;i++)
    targetNode[jobPos][i] = i % num_task_tracker;
  for (i=0;i<MAX_TASK_PER_JOB;i++)
    targetNode[jobPos][i] = taskTrackerLocalToGlobal(targetNode[jobPos][i]);
}

void scheduleReduceTask( tw_lpid * targetNode,
			 int * numNode )
{
  // scheduling algorithm is complicated,
  // leave it for now
  int i;
  *numNode = NUM_TASK_PER_JOB;
  for (i=0;i<*numNode;i++)
    *(targetNode+i) = i % num_task_tracker;
}

void outputSlotMap( int * slotMap )
{
  int i;
  printf("\n\t Now slotMap is :\n");
  for (i=0;i<num_task_tracker;i++)
    printf(" %d",slotMap[i]);
  printf("\n");
}

void outputScheduleMap( tw_lpid ** Map )
{
  int i,j;
  printf("\n\t Now Schedule Map is :\n\n");
  for (j=0;j<MAX_NUM_JOB;j++)
    {
    for (i=0;i<MAX_TASK_PER_JOB;i++)
      {
	printf(" %d",Map[j][i]);
      }
    printf("\n\n");
    }
}

int taskTrackerGlobalToLocal(int gid)
{
  int lid, base, grp_id;
  // dragonfly specific
  base = num_tt_per_group + num_dn_per_group + num_tm_per_group + 1;
  grp_id = gid/base;
  // this one is tricky, because task_tracker is mapped at the
  // very beginning, otherwise, you should add other lps
  lid = grp_id * num_tt_per_group + gid % base;
  return lid;
}

int taskTrackerLocalToGlobal(int lid)
{
  int gid, base, grp_id;
  // dragonfly specific
  base = num_tt_per_group + num_dn_per_group + num_tm_per_group + 1;
  grp_id = lid/num_tt_per_group;
  //printf("base is %d and group_id is %d and lid is %d\n",base,grp_id,lid);
  // this one is tricky, because task_tracker is mapped at the
  // very beginning, otherwise, you should add other lps
  gid = grp_id * base + lid % num_tt_per_group;
  return gid;
}

void task_tracker_init(task_tracker_state* ns,  tw_lp* lp)
{

    printf("TaskTraker report, lp id is %d\n",lp->gid);

    tw_event *e;
    mr_msg *m;
    tw_stime kickoff_time;

    // initialize job tracker
    int i, j;
    if (lp->gid == JOB_TRACKER_POS) // job tracker
      {
	ns->jt_state = (job_tracker_state *)malloc(sizeof(job_tracker_state));
	// actually, the first task tracker is job tracker, but we assign more space
	(ns->jt_state)->slotMap = (int *)malloc(sizeof(int)*num_task_tracker);
	for (i=0;i<num_task_tracker;i++)
	  (ns->jt_state)->slotMap[i] = MAX_NUM_SLOT;

	(ns->jt_state)->job_schedule = (tw_lpid **)malloc(sizeof( tw_lpid *) * MAX_NUM_JOB);
	for (i=0;i<MAX_NUM_JOB;i++)
	  (ns->jt_state)->job_schedule[i] = (tw_lpid *)malloc(sizeof( tw_lpid ) * MAX_TASK_PER_JOB);

	for (i=0;i<MAX_NUM_JOB;i++)
	  for (j=0;j<MAX_TASK_PER_JOB;j++)
	    (ns->jt_state)->job_schedule[i][j] = -1;

	//outputScheduleMap((ns->jt_state)->job_schedule);

	// initialize a job, this should be read in or configured in future
	for (i=0;i<MAX_NUM_JOB;i++)
	  {
	    (ns->jt_state)->mr_job[i].range_start = 0;
	    (ns->jt_state)->mr_job[i].range_start = 128*1024*1024;
	    (ns->jt_state)->mr_job[i].num_splits = 3;
	  }	  

	(ns->jt_state)->map_task_finish_count = 0;

      }
    else // normal task tracker
      {
	ns->jt_state = NULL;
      }
    
    // initiate heart beat
    if (lp->gid != JOB_TRACKER_POS) // task tracker
      {
	// local event
	kickoff_time = g_tw_lookahead + tw_rand_unif(lp->rng); 

	e = codes_event_new(lp->gid, kickoff_time, lp);
	m = tw_event_data(e);
	m->mr_event_type = MR_HEART_BEAT;
	//tw_event_send(e);
      }


    //printf("lookahead is %lf\n",g_tw_lookahead);

    kickoff_time = g_tw_lookahead + tw_rand_unif(lp->rng); 

    e = codes_event_new(lp->gid, kickoff_time, lp);
    m = tw_event_data(e);
    m->mr_event_type = MR_KICKOFF;
    tw_event_send(e);

    ns->reduce_task_finish_count = 0;
    ns->counter_1 = 0;

    return;
}

void task_tracker_eventHandler(task_tracker_state* ns, tw_bf* b, mr_msg* m, tw_lp* lp)
{
  switch (m->mr_event_type)
    {
    case MR_GET_INTER_DATA_BACK:
      handle_get_inter_data_back_event(ns, b, m, lp);
      break;
    case MR_START_MAP_TASK:
      handle_start_map_task_event(ns, b, m, lp);
      break;
    case MR_FINISH_MAP_TASK:
      handle_finish_map_task_event(ns, b, m, lp);
      break;
    case MR_START_REDUCE_TASK:
      handle_start_reduce_task_event(ns, b, m, lp);
      break;
    case MR_FINISH_REDUCE_TASK:
      handle_finish_reduce_task_event(ns, b, m, lp);
      break;
    case MR_GET_DATA_BLOCK_BACK:
      handle_get_data_block_back_event(ns, b, m, lp);
      break;
    case MR_REDUCER_COMPUTE:
      handle_reducer_compute_event(ns, b, m, lp);
      break;
    case MR_GET_DATA_BLOCK:
      handle_get_data_block_event(ns, b, m, lp);
      break;
    case MR_ACK_BLOCK_POS_QUERY:
      handle_ack_block_pos_query_event(ns, b, m, lp);
      break;
    case MR_BLOCK_POS_QUERY:
      handle_block_pos_query_event(ns, b, m, lp);
      break;
    case MR_RECV_MAP_TASK:
      handle_recv_map_task_event(ns, b, m, lp);
      break;
    case MR_ASSIGN_MAP_TASK:
      handle_assign_map_task_event(ns, b, m, lp);
      break;
    case MR_SCHEDULE:
      handle_schedule_event(ns, b, m, lp);
      break;
    case MR_HEART_BEAT:
      handle_heartbeat_event(ns, b, m, lp);
      break;
    case MR_RECV_HEART_BEAT:
      handle_heartbeat_recv_event(ns, b, m, lp);
      break;
    case MR_REQ:
      handle_req_event(ns, b, m, lp);
      break;
    case MR_ACK:
      handle_ack_event(ns, b, m, lp);
      break;
    case MR_KICKOFF:
      handle_kickoff_event(ns, b, m, lp);
      break;
    case MR_LOCAL:
      handle_local_event(ns, b, m, lp); 
      break;
    default:
      printf("\n Invalid message type %d ", m->mr_event_type);
      printf("\n lp %d speaking and msg src is %d\n", lp->gid, m->src);
      exit(0);
      //assert(0);
      break;
    }
}

void task_tracker_eventHandler_rc(task_tracker_state* ns, tw_bf* b, mr_msg* m, tw_lp* lp)
{
  switch (m->mr_event_type)
    {
    case MR_GET_INTER_DATA_BACK:
      handle_get_inter_data_back_event_rc(ns, b, m, lp);
      break;
    case MR_START_MAP_TASK:
      handle_start_map_task_event_rc(ns, b, m, lp);
      break;
    case MR_FINISH_MAP_TASK:
      handle_finish_map_task_event_rc(ns, b, m, lp);
      break;
    case MR_GET_DATA_BLOCK_BACK:
      handle_get_data_block_back_event_rc(ns, b, m, lp);
      break;
    case MR_REDUCER_COMPUTE:
      handle_reducer_compute_event_rc(ns, b, m, lp);
      break;
    case MR_START_REDUCE_TASK:
      handle_start_reduce_task_event_rc(ns, b, m, lp);
      break;
    case MR_GET_DATA_BLOCK:
      handle_get_data_block_event_rc(ns, b, m, lp);
      break;
    case MR_ACK_BLOCK_POS_QUERY:
      handle_ack_block_pos_query_event_rc(ns, b, m, lp);
      break;
    case MR_BLOCK_POS_QUERY:
      handle_block_pos_query_event_rc(ns, b, m, lp);
      break;
    case MR_RECV_MAP_TASK:
      handle_recv_map_task_event_rc(ns, b, m, lp);
      break;
    case MR_ASSIGN_MAP_TASK:
      handle_assign_map_task_event_rc(ns, b, m, lp);
      break;
    case MR_SCHEDULE:
      handle_schedule_event_rc(ns, b, m, lp);
      break;
    case MR_HEART_BEAT:
      handle_heartbeat_event_rc(ns, b, m, lp);
      break;
    case MR_RECV_HEART_BEAT:
      handle_heartbeat_recv_event_rc(ns, b, m, lp);
      break;
    case MR_REQ:
      handle_req_rev_event(ns, b, m, lp);
      break;
    case MR_ACK:
      handle_ack_rev_event(ns, b, m, lp);
      break;
    case MR_KICKOFF:
      handle_kickoff_rev_event(ns, b, m, lp);
      break;
    case MR_LOCAL:
      handle_local_rev_event(ns, b, m, lp);    
      break;
    default:
      assert(0);
      break;
    }

    return;
}

/* convert ns to seconds */
static tw_stime ns_to_s(tw_stime ns)
{
    return(ns / (1000.0 * 1000.0 * 1000.0));
}

/* convert seconds to ns */
static tw_stime s_to_ns(tw_stime ns)
{
    return(ns * (1000.0 * 1000.0 * 1000.0));
}

void task_tracker_finish(task_tracker_state* ns, tw_lp* lp)
{

  int i;
  
  if (lp->gid == JOB_TRACKER_POS) // job tracker
    {
      /* free((ns->jt_state)->slotMap); */
      /* free(ns->jt_state); */
      /* for (i=0;i<MAX_NUM_JOB;i++) */
      /* 	free( (ns->jt_state)->job_schedule[i] ); */
      /* free((ns->jt_state)->job_schedule); */
    }
  
    //printf("Task Tracker %d finish\n",lp->gid);
  /* printf("server %llu recvd %d bytes in %f seconds, %f MiB/s sent_count %d recvd_count %d local_count %d \n", (unsigned long long)lp->gid, PAYLOAD_SZ*ns->msg_recvd_count, ns_to_s((tw_now(lp)-ns->start_ts)),  */
  /*       ((double)(PAYLOAD_SZ*NUM_REQS)/(double)(1024*1024)/ns_to_s(tw_now(lp)-ns->start_ts)), ns->msg_sent_count, ns->msg_recvd_count, ns->local_recvd_count); */
  return;
}



